// todo требовать логгер при подключении или пробрасывать конфиг для нового логгера
const amqp = require('amqplib');
const _ = require('lodash');

const Logger = require('./Logger');
const ErrorWrp = require('./Error');

const SAError = new ErrorWrp('AMQP');

/**
 * @typedef Config
 * @property uri {string} Connection string
 * @property queues {Queue[]} List of queues options
 */

/**
 * @typedef Queue
 * @property name {string} Name of queue
 * @property channelOptions {ChannelOptions} Options for queue's channel
 * @property assertOptions {AssertOptions} Options for assert queue
 */

/**
 * @typedef ChannelOptions
 * @property prefetch {number} messages per work
 */

/**
 * @typedef AssertOptions
 * @property durable {boolean} Is durable?
 */

/**
 * @typedef ConsumeOptions
 * @property noAck {boolean} if false consumer must ack message by yourself
 */

class AMQP {
  static async connect(config, parentLogger = {}) {
    const amqpInstance = new AMQP(config, parentLogger);
    await amqpInstance.connect();

    return amqpInstance;
  }

  /**
   * Creates an instance of AMQP.
   * @param {Config} config
   * @param {*} parentLogger
   * @memberof AMQP
   */
  constructor(config, parentLogger) {
    this.logger = parentLogger.copy({ uptime: Date.now() }, Logger.TYPE.AMQP) || new Logger({ uptime: Date.now() }, Logger.TYPE.AMQP);

    /**
     * @type Config
     */
    this.config = {
      uri: 'amqp://localhost',
      queues: []
    };

    _.merge(this.config, config);

    if (!this.config.queues
       || !Array.isArray(this.config.queues)
       || this.config.queues.length === 0) {
      SAError.throw(new Error('Config.queues must be not empty array'), 500, this.config);
    }

    this.requestsCount = 0;

    this.connection = null;

    this.queues = {};

    for (const queueObj of this.config.queues) {
      if (this.queues[queueObj.name]) {
        SAError.throw(new Error('duplicate queues not allowed'), 500, this.config);
      }

      this.queues[queueObj.name] = _.cloneDeep(queueObj);
    }
  }

  async connect() {
    this.connection = await amqp.connect(this.config.uri);

    for (const queueName in this.queues) {
      if (this.queues.hasOwnProperty(queueName)) {
        const queueObj = this.queues[queueName];

        const channel = await this.connection.createChannel();

        if (queueObj.channelOptions && queueObj.channelOptions.prefetch) {
          channel.prefetch(queueObj.channelOptions.prefetch);
        }

        await channel.assertQueue(queueName, queueObj.assertOptions);

        queueObj.channel = channel;
      }
    }
  }

  async end() {
    for (const queueName in this.queues) {
      if (this.queues.hasOwnProperty(queueName)) {
        const queueObj = this.queues[queueName];
        queueObj.channel.close();
      }
    }

    await this.connection.close();
  }

  /**
   * Send message
   *
   * @param {string} queueName
   * @param {string|{}|Buffer} message
   * @returns Promise<{}|void>
   * @memberof AMQP
   */
  async sendMessageToQueue(queueName, message) {
    if (!this.queues[queueName]) {
      SAError.throw(new Error(`Queue with name=${queueName} is not exists`));
    }

    let messageForSending = null;

    if (typeof message === 'string') {
      messageForSending = Buffer.from(message);
    } else if (typeof message === 'object') {
      messageForSending = Buffer.from(JSON.stringify(message));
    } else if (Buffer.isBuffer(message)) {
      messageForSending = message;
    }

    try {
      return await this.queues[queueName].channel.sendToQueue(queueName, messageForSending);
    } catch (e) {
      SAError.throw(e, 500, this.config);

      return null;
    }
  }

  /**
   * Consume to messages from queue
   *
   * @param {string} queueName
   * @param {function} handler
   * @param {ConsumeOptions} consumeOptions
   * @returns Promise<{}|void>
   * @memberof AMQP
   */
  async listenQueue(queueName, handler, consumeOptions = {}) {
    const channel = this.queues[queueName].channel;
    await channel.consume(queueName, (msg) => {
      try {
        handler(msg, {
          channel
        });
      } catch (e) {
        SAError.throw(e, 500, this.config);
      }
    }, consumeOptions);
  }
}

module.exports = AMQP;
