/**
 * Created by Nikolay Ilin (na.iline@me.com) on 18.10.2017.
 * Copyrights by Soft-Artel LIMITED
 * www.soft-artel.com
 */

require('sa-prototype');

const fs = require('./FsAsync');
const Pgsql = require('./Pgsql');
const Redis = require('./Redis');
const AMQP = require('./AMQP');

let logger;

async function connect(initSate, config, applogger, endConnection) {
  logger = applogger;

  logger.info('Connect to dependencies...');

  if (config.pgsqlM && !initSate.dbM) {
    initSate.dbM = await Pgsql.connect(config.pgsqlM, logger, endConnection);
  }

  if (config.pgsqlS && !initSate.dbS) {
    initSate.dbS = await Pgsql.connect(config.pgsqlS, logger, endConnection);
  }

  if (config.redis && !initSate.cache) {
    initSate.cache = await Redis.connect(config.redis, logger);
  }

  if (config.amqp && !initSate.queue) {
    initSate.queue = await AMQP.connect(config.amqp, logger);
  }

  if (config.storage) {
    try {
      await fs.access(config.storage, fs.constants.W_OK);
      initSate.storage = config.storage;
    } catch (err) {
      if (err.code !== 'EEXIST') {
        err.status = 500;
        err.params = { storagePath: config.storage };
        throw err;
      }
    }
  }

  const checkRes = await check(initSate);

  if (checkRes.status !== 1) {
    const err = new Error('Dependencies self check FAIL');
    err.status = 500;
    err.params = checkRes;
    throw err;
  }

  return checkRes;
}

async function end(state) {
  const stateKeys = Object.keys(state);

  for (const key of stateKeys) {
    if (!state[key].end || typeof state[key].end !== 'function') {
      continue;
    }

    try {
      logger.info(`Closing ${key} connection...`);
      await state[key].end();
    } catch (err) {
      logger.error(err);
    }
  }
}

async function check(state, reqErrors) {
  const result = logger.appInfo;

  result.status = 0;
  result.errors = [];
  result.deps = [];

  if (state.dbM) {
    try {
      const res = await state.dbM.runSQL(state.dbM.config.test);

      if (res) {
        result.deps.push('MASTER Pgsql: OK');
        // logger.info('MASTER Pgsql is OK!', { processID: state.dbM.db.processID });
      } else {
        result.deps.push('MASTER Pgsql: WARN');
        logger.error('MASTER Pgsql selfCheck WARN', res);
      }
    } catch (err) {
      result.errors.push(err.message);
      result.deps.push('MASTER Pgsql: ERR');
      logger.error(err);
    }
  }

  if (state.dbS) {
    try {
      const res = await state.dbS.runSQL(state.dbS.config.test);

      if (res) {
        result.deps.push('SLAVE Pgsql: OK');
        // logger.info('SLAVE Pgsql is OK!', { processID: state.dbS.db.processID });
      } else {
        result.deps.push('SLAVE Pgsql: WARN');
        logger.error('SLAVE Pgsql selfCheck WARN', res);
      }
    } catch (err) {
      result.errors.push(err.message);
      result.deps.push('SLAVE Pgsql: ERR');
      logger.error(err);
    }
  }

  if (state.cache) {
    try {
      await state.cache.setex(`${result.app}-selfCheck`, 'PASS', 10);
      const val = await state.cache.get(`${result.app}-selfCheck`);

      if (val === 'PASS') {
        result.deps.push('Redis: OK');
        // logger.info('Redis is OK!');
      } else {
        result.deps.push('Redis: WARN');
        logger.error('Redis selfCheck WARN', { val });
      }
    } catch (err) {
      result.errors.push(err.message);
      result.deps.push('Redis: ERR');
      logger.error(err);
    }
  }

  if (state.storage) {
    try {
      const filePath = `${state.storage + result.app}-check.txt`;

      await fs.writeFile(filePath, 'PASS', 'utf-8');
      const val = await fs.readFile(filePath, 'utf-8');

      if (val === 'PASS') {
        result.deps.push('Storage: OK');
        // logger.info('Storage is OK!');
      } else {
        result.deps.push('Storage: WARN');
        logger.error('Storage selfCheck WARN', { val });
      }

      await fs.unlink(filePath);
    } catch (err) {
      result.errors.push(err.message);
      result.deps.push('Storage: ERR');
      logger.error(err, { storagePath: state.storage });
    }
  }

  if (result.errors.length === 0) {
    result.status = 1;
    // logger.info('Self check is OK...', reqErrors);
  }

  if (reqErrors) {
    result.errors.push(JSON.stringify(reqErrors));
  }

  return result;
}

module.exports = { connect, check, end };
