// todo требовать логгер при подключении или пробрасывать конфиг для нового логгера
/**
 * Created by Asha (axaxat@gmail.com) on 15.08.2018.
 * Copyrights by Soft-Artel LIMITED
 * www.soft-artel.com
 */

require('sa-prototype');

const sqlHelper = require('sqlstring');
const pg = require('pg');

const { Pool } = pg;

const Logger = require('./Logger');
const ErrorWrp = require('./Error');

const Error = new ErrorWrp('PGSql');

const types = pg.types;

types.setTypeParser(1114, (stringValue) => new Date(`${stringValue}Z`));

module.exports = class Pgsql {
  static async connect(config, parentLogger = {}, endConnection) {
    const db = new Pgsql(config, parentLogger);
    await db.connect(endConnection);

    return db;
  }

  // -----------------------------------------------

  constructor(config, parentLogger) {
    this.logger = parentLogger.copy({ uptime: Date.now() }, Logger.TYPE.DB) || new Logger({ uptime: Date.now() }, Logger.TYPE.DB);

    // https://node-postgres.com/api/client
    this.config = {
      statement_timeout: 0,
      host: 'localhost',
      port: 5432,
      test: 'select timeofday()',
      search_path: 'sp2,public',
      idleTimeoutMillis: 0,
      connectionTimeoutMillis: 600000
    };

    this.config.mergeFrom(config);
    this.format = sqlHelper.format;
    this.state = 0;
  }

  // -----------------------------------------------

  async connect(endConnection) {
    try {
      this.db = new Pool(this.config);
      // eslint-disable-next-line no-unused-vars
      this.db.on('error', (err, client) => {
        this.logger.error(err, { code: err.code }, true);
        endConnection(err);
      });

      this.requestsCount = 0;

      this.logger.debug('PGsql connected. ');
    } catch (err) {
      Error.throw(err, 500, this.config);
    }
  }

  // -----------------------------------------------

  async end() {
    try {
      await this.db.end();
      this.logger.debug(`PGsql connection closed, ${this.requestsCount} requests complete`);
    } catch (err) {
      Error.throw(err, 500, this.config);
    }
  }
  // -----------------------------------------------

  async list(Model, {
    sql,
    where,
    ID,
    userID,
    order,
    asModel = false,
    groupResultBy = null,
    groupResultArrays = false
  },
  context, logger, skipAclCheck = false) {
    if (!Model) {
      Error.throw('list: Model is undefined!');
    }

    if (!where && !ID && !sql) {
      Error.throw('list: Need ID or where or sql to select!');
    }

    const options = { table: Model.table };

    if (sql) {
      options.sql = sql;
    }

    if (ID && !where) {
      where = { ID };
    }

    if (where && userID) {
      where.userID = userID;
    }

    if (where) {
      options.where = where;
    }

    if (asModel && context) {
      options.Model = Model;
      options.ModelContext = context;
    }

    options.order = order;
    options.groupResultBy = groupResultBy;
    options.groupResultArrays = groupResultArrays;

    return await this.select(options, logger || context ? context.logger : null, skipAclCheck) || [];
  }

  // -----------------------------------------------

  async select({
    sql = null, // Ready SQL
    params = null, // PGsql.format ? params inject to SQL
    // OR Build SQL By:
    fields = '*', // Fields in result in double quotes "ID"
    table = null, // Table Name
    where = null, // Where STRING or OBJ { field: value }
    order = null, // STRING Order By
    limit = null, // Limit, if == 1, result = {<row>}

    groupResultBy = null, // Field to group result
    groupResultArrays = false, // If true, result = { ${groupResultBy}: [<rows>] }.

    Model = null, // Model to result
    ModelContext = null,
    streamHandler = null
  }, parentLogger, skipAclCheck = false) {
    if (Model && !Model.createFromStoredFields) {
      Error.throw('select: WRONG MODEL - NO createFromStoredFields() function!');
    }

    if (Model && !ModelContext) {
      Error.throw('select: NO ModelContext, but Model exist');
    }

    // eslint-disable-next-line prefer-rest-params
    if (typeof arguments['0'] === 'string') {
      // eslint-disable-next-line prefer-rest-params
      sql = arguments['0'];
    }

    if (sql && params) sql = this.format(sql, params);

    if (!sql) sql = makeSelectSQL(table, where, limit, fields, order);

    // stream

    if (streamHandler) {
      return await this.runSQL(sql, streamHandler, 0, parentLogger);
    }

    // Counters and Metrics

    const [rows] = await this.runSQL(sql, null, 0, parentLogger);

    if (limit === 1 && rows.length === 0) {
      return undefined;
    }

    if (limit === 1 && rows.length === 1 && !groupResultBy) {
      return Model ? Model.createFromStoredFields(rows[0], ModelContext, skipAclCheck) : rows[0];
    }

    if (!groupResultBy && !Model) {
      return rows;
    }

    const modelRows = [];
    const res = {};

    for (let row of rows) {
      if (Model && !groupResultBy) {
        modelRows.push(Model.createFromStoredFields(row, ModelContext, skipAclCheck));
        continue;
      }

      const key = row[groupResultBy];

      if (!key) {
        Error.throw(`select: NO field [${groupResultBy}] in select result`, 500, row);
      }

      if (Model) {
        row = Model.createFromStoredFields(row, ModelContext, skipAclCheck);
      }

      if (!res[key] && groupResultArrays) res[key] = [];

      if (groupResultArrays) {
        res[key].push(row);
      } else {
        res[key] = row;
      }
    }

    return Model && !groupResultBy ? modelRows : res;
  }

  /**
   @param {String} table
   @param {Array|Object} data
   @param {Object|null} where
   @return {Number} Affected rows
   */
  async update(table, data, where = null) {
    // eslint-disable-next-line no-unused-vars
    const [results, rowCount] = await this.runSQL(makeUpdDbSQL(table, data, 'update', where));

    return rowCount;
  }

  /**
   @param {String} table
   @param {Array|Object} data
   @return {Number} Autoincrement Id or Affected rows
   */
  async insert(table, data) {
    const [results, rowCount] = await this.runSQL(makeUpdDbSQL(table, data, 'insert'));

    // TODO returning *
    if (results && results.insertId) return results.insertId;

    return rowCount;
  }

  /**
   @param {String} table
   @param {Object} data
   @param {Array}  keys to update if insert is dublicate
   @return {Number} Autoincrement Id or Affected rows
   */
  async insertOrUpdate(table, data, insUpdCols = null) {
    const [results, rowCount] = await this.runSQL(makeUpdDbSQL(table, data, 'insert|update', null, insUpdCols));

    if (results && results.insertId) return results.insertId;

    if (!results || !rowCount) {
      return 'FAIL';
    }

    if (rowCount === 0) {
      return undefined;
    }

    // FIXME
    if (rowCount === 1) {
      return 'INSERT';
    }

    // FIXME
    if (rowCount === 2) {
      return 'UPDATE';
    }

    return results;
  }

  /**
   @param {String} table
   @param {Array|Object} data
   @param {Bool} forceDelete
   @return {Number} Affected rows
   */

  async delete(table, where, forceDelete = false) {
    if (!where && !forceDelete) {
      Error.throw(`Delete from ${table} without WHERE!`, 500);
    }

    // eslint-disable-next-line no-unused-vars
    const [results, rowCount] = await this.runSQL(makeUpdDbSQL(table, null, 'delete', where));

    return rowCount;
  }

  // eslint-disable-next-line no-unused-vars
  async stream(sql = '', processRow = null, withPauses = true) {
    Error.throw('Not implemented');
  }

  async query(sql, params) {
    return await this.runSQL(this.format(sql, params));
  }

  // eslint-disable-next-line no-unused-vars
  makeMultyUpdDbSQL(table, data, type = 'insert', where = null) {
    return makeMultyInsertDbSQL(table, data);
  }

  // eslint-disable-next-line consistent-return
  async runSQL(sql, streamHandler, recursionCnt = 0, parentLogger) { // eslint-disable-line no-unused-vars
    this.requestsCount++;

    // const logger = parentLogger ? parentLogger.copy({}, this.logger.type) : this.logger;

    // logger.dump(`PGsql query # ${this.requestsCount}`, { sql });

    let queryTime = Date.now();

    try {
      this._lastSQL = sql;

      const res = streamHandler ? await this.stream(sql, streamHandler) : await this.db.query(sql);

      queryTime = Date.now() - queryTime;

      // logger.debug(`PGsql query # ${this.requestsCount} complete by ${String.humanizeMsDuration(queryTime)}`, { rowCount: res.rowCount });

      return [res.rows, res.rowCount];
    } catch (err) {
      Error.throw(err, 500, { req: this.requestsCount, time: String.humanizeMsDuration(queryTime), sql });
    }
  }
};

function makeSelectSQL(table = null, where = null, limit = null, cols = '*', order = null) {
  const values = [];

  const params = [];

  let sql = `SELECT ${cols} FROM ${table}`;

  if (where) {
    sql += ' WHERE ';

    if (typeof where === 'string') {
      sql += ` ${where}`;
    } else if (typeof where !== 'object') { // FIXME не понятное условие. Что значит массив params?
      sql += ' ?';
      params.push(where);
    } else {
      Object.keys(where).forEach((key) => {
        // IN STATEMENT
        if (Array.isArray(where[key]) && where[key].length > 0) {
          let sqlIn = '';

          where[key].forEach((val) => {
            sqlIn += ` $${values.push(val)}, `;
          });

          sqlIn = sqlIn.slice(0, -2);

          sql += ` "${key.replace(/[='"]/g, '')}" IN ( ${sqlIn} ) AND `;
        }

        if (Array.isArray(where[key]) && where[key].length === 0) {
          sql += ` "${key.replace(/[='"]/g, '')}" = NULL AND `;
        }

        if (!Array.isArray(where[key])) {
          sql += ` "${key.replace(/[='"]/g, '')}" = $${values.push(where[key])} AND `;
        }
      });

      sql += ' TRUE';
    }
  }

  if (order) {
    sql += ` ORDER BY ${order}`;
  }

  if (limit) {
    sql += ` LIMIT ${limit}`;
  }

  sql += ' ;';

  return {
    text: sql,
    values
  };
}

function makeUpdDbSQL(table, data, type = 'insert', where = null, insUpdCols = null) {
  if (Array.isArray(data) && data.length === 1) {
    data = data[0];
  }

  if (Array.isArray(data) && (type === 'insert' || type === 'replace')) {
    return makeMultyInsertDbSQL(table, data, type);
  }

  let sql = '';

  switch (type) {
    case 'update':
      sql = `UPDATE ${table} SET `;
      break;

    case 'insert':
      sql = `INSERT INTO  ${table} `;
      break;

    case 'replace':
      sql = `REPLACE ${table} SET ?`;
      Error.throw(`Not implemented:${sql}`);
      break;

    case 'insert|update':
      sql = `INSERT INTO ${table} `;
      break;

    case 'delete':
      sql = `DELETE FROM ${table}`;
      break;
  }

  const values = [];

  // delete empty strings
  for (const [field, value] of Object.entries(data)) {
    if (typeof value !== 'object' && value === ' ') {
      delete data[field];
    }
  }

  const cols = Object.keys(data);

  if (type === 'update') {
    sql += `( "${cols.join('", "')}") = ( `;

    cols.forEach((col) => {
      if (data[col] && typeof data[col] === 'object') {
        sql += `$${values.push(JSON.stringify(data[col]))}, `;
      } else {
        sql += `$${values.push(data[col])}, `;
      }
    });

    sql = sql.substr(0, sql.length - 2);
    sql += ')';
  }

  if (type === 'insert' || type === 'insert|update') {
    sql += `( "${cols.join('", "')}") VALUES( `;

    cols.forEach((col) => {
      if (data[col] && typeof data[col] === 'object') {
        sql += `$${values.push(JSON.stringify(data[col]))}, `;
      } else {
        sql += `$${values.push(data[col])}, `;
      }
    });

    sql = sql.substr(0, sql.length - 2);
    sql += ')';
  }

  if (type === 'insert|update') {
    sql += ` ON CONFLICT ON CONSTRAINT ${table}_pkey DO UPDATE SET `;

    if (!insUpdCols) {
      insUpdCols = Object.keys(data);
    }

    // data[col] can be undefined see 429 line
    insUpdCols.forEach((col) => {
      if (typeof data[col] !== 'undefined') {
        if (data[col] && typeof data[col] === 'object') {
          sql += ` "${col}" = $${values.push(JSON.stringify(data[col]))}, `;
        } else {
          sql += ` "${col}" = $${values.push(data[col])}, `;
        }
      }
    });

    sql = sql.substr(0, sql.length - 2);
  }

  if (where) {
    sql += ' WHERE ';

    if (typeof where === 'string') {
      sql += where;
    } else {
      Object.keys(where).forEach((field) => {
        sql += ` "${field}" =  $${values.push(where[field])} AND `;
      });

      sql += ' TRUE';
    }
  }

  sql += ';';

  return {
    text: sql,
    values
  };
}

function makeMultyInsertDbSQL(table, data, type = 'insert') {
  if (type !== 'insert' && type !== 'replace') return false;

  // В PG нет replace into
  if (type === 'replace') {
    Error.throw('Not implemented:  REPLACE');
  }

  const keys = Object.keys(data[0]);

  if (!keys || keys.length === 0) {
    return null;
  }

  let sql = `${type.toUpperCase()} INTO ${table} ("${keys.join('", "')}") VALUES `;
  const valuesRows = [];
  data.forEach((row) => {
    sql += '(';

    keys.forEach((key) => {
      if (typeof row[key] === 'object' && !Buffer.isBuffer(row[key])) {
        sql += `$${valuesRows.push(JSON.stringify(row[key]))}, `;
      } else {
        sql += `$${valuesRows.push(row[key])}, `;
      }
    });

    sql = sql.substr(0, sql.length - 2);
    sql += '),';
  });

  sql = sql.substr(0, sql.length - 1);
  sql += ';';

  return {
    text: sql,
    values: valuesRows
  };
}
