/**
 * Created by Nikolay Ilin (na.iline@me.com) on 14.12.2017.
 * Copyrights by Soft-Artel LIMITED
 * www.soft-artel.com
 */

require('sa-prototype');

const bodyParse = require('co-body');
const path = require('path');
const bytes = require('bytes');
const mime = require('mime-types');
const Koa = require('koa');

const helpers = require('../helpers');
const { validateFields } = require('../Validate');
const fs = require('../FsAsync');
const Logger = require('../Logger');
const deps = require('../Dependencies');
const ErrorWrp = require('../Error');
const authFieldsInitFun = require('./authFieldsInit');

/**
 * Create server instance with all needed dependencies
 *
 * @param {string} dirname Main directory for app
 * @returns {{start: function, end: function}}
 */
const getServer = (dirname) => {
  const pkg = require(`${dirname}/app.json`);

  const REQUEST_TIMEOUT = 0;

  const SAError = new ErrorWrp('Server');

  const app = new Koa();

  let applogger = null;

  const reqErrors = {};

  const config = {
    port: 80,
    path: {
      middlewares: `${dirname}/middlewares/`,
      commands: `${dirname}/commands/`
    },
    response: {
      json: true,
      errors: true
    }
  };

  const contract = {
    version: 'v1'
  };

  const staticExt = ['.txt', '.html', '.js', '.css', '.jpg', '.jpeg', '.gif',
    '.ico', '.png', '.svg', '.otf', '.ttf', '.zip', '.map',
    '.safariextz', '.woff', '.pdf', '.dmg', '.docx', '.exelx', '.json'];

  const jsModules = {};
  const reqRules = {};
  const defaultCmd = {};

  let initSate = {};

  /**
   * Start application
   *
   * @async
   * @param {{}} [appConfig={}]
   * @param {{}} [appContract={}]
   * @param {{}} [appState={}]
   * @returns {Koa|void}
   */
  // eslint-disable-next-line consistent-return
  async function start(appConfig = {}, appContract = {}, appState = {}) {
    try {
      config.mergeFrom(appConfig);
      contract.mergeFrom(appContract);
      initSate = appState;

      applogger = new Logger(
        {
          app: pkg.name,
          ver: pkg.version,
          stage: config.stage,
        },
        Logger.TYPE.APPLICATION,
        config.loggerOptions || {}
      );

      applogger.debug('starting...');

      await deps.connect(initSate, config, applogger, end);

      await loadContract();

      const authFieldsInit = authFieldsInitFun(config);
      // Сервисный мидлвар - для каждого запроса самый первый

      applogger.info('Add Server middleware...');

      app.use(async (ctx, next) => {
        let reqlogger = applogger;
        let reqError = null;

        try {
          // -----------------------------------------------------------------------------------------------
          // 1. Предустановленные локейшены

          if (ctx.path.endsWith('_healthz')) {
            ctx.body = await deps.check(initSate, reqErrors);
            ctx.status = ctx.body.status === 1 ? 200 : 503;

            return;
          }

          if (ctx.path.endsWith('apppParseddContract')) {
            ctx.body = contract;
            ctx.status = 200;

            return;
          }

          if (ctx.path.endsWith('robots.txt')) {
            ctx.body = 'User-agent: *\nDisallow: /';
            ctx.status = 200;

            return;
          }

          if (ctx.path.endsWith('favicon.ico') && !config.path.static) {
            ctx.status = 404;

            return;
          }

          // -----------------------------------------------------------------------------------------------
          // 2. отдаем статику

          if (config.path.static && staticExt.indexOf(path.extname(ctx.path)) !== -1) {
            let filePath = config.path.static + ctx.path;
            filePath = filePath.replace('//', '/'); // path.normalize(filePath)

            try {
              ctx.type = mime.lookup(filePath);
              ctx.body = await fs.readFile(filePath); // ?
              ctx.status = 200;
            } catch (err) {
              this.status = 404;
            }

            return;
          }

          // -----------------------------------------------------------------------------------------------
          // Нормальный реквест - создаем логгер

          reqlogger = applogger.copy({ rts: Date.now() }, Logger.TYPE.REQUEST);
          helpers.updateLoggerEnvByReqCtx(ctx, reqlogger);

          // -----------------------------------------------------------------------------------------------
          // 3. Вычисляем команду

          let location = ctx.path; //   /v1.0/location/[location/]command
          let command;

          if (contract.version) {
            location = location.replace(contract.version, '');
          }

          if (contract.skipUrlAppPath) {
            let appName = pkg.name.split('-');
            appName = appName.length > 1 ? appName[1] : pkg.name;
            location = location.replace(appName, '');

            reqlogger.debug('appName', appName);
          }

          // eslint-disable-next-line no-useless-escape
          location = location.replace(/[\/]{2,}/g, '/'); //  //->/
          location = location.length > 1 ? location.replace(/\/$/, '') : location; // /loc/ => /loc

          location = defaultCmd[location] || location;

          // eslint-disable-next-line no-useless-escape
          command = location.match(/[^\/]+$/ig);
          command = command && command.length > 0 ? command[0] : 'none';

          ctx.state.command = command;
          ctx.state.location = location;

          reqlogger.setEnv('cmd', command);

          const ErrorReq = new ErrorWrp(`${ctx.method} ${location}`);

          // -----------------------------------------------------------------------------------------------
          // 4. Начинаем обработку запроса

          reqlogger.info(`-> ${ctx.method.toUpperCase()} ${ctx.state.location || ctx.originalUrl}`);

          if (contract.commands) {
            if (reqRules[location]) {
              ctx.state.reqRule = reqRules[location] || {};
            } else {
              SAError.throw(`Wrong location: ${location}`, 400, Object.keys(reqRules));
            }
          } else {
            ctx.state.reqRule = contract.reqRules || {};
          }

          // -----------------------------------------------------------------------------------------------
          // 5. Выставляем заголовки ответа
          /**
           * запросы типа OPTION обрабатываются nginx и заголовки выставляются nginx,
           */
          if (ctx.method === 'OPTIONS') {
            ctx.status = 200;
            ctx.body = 'OK';

            return;
          }

          // -----------------------------------------------------------------------------------------------
          // 6. Валидация запроса

          const reqRule = ctx.state.reqRule;

          reqlogger.debug('Validate request...');

          // reqRule
          // 6.1. Проверяем метод

          if (reqRule.reqMethod && ctx.method.toUpperCase() !== reqRule.reqMethod.toUpperCase()) {
            ErrorReq.throw(`Method ${ctx.method} NOT Allowed for location:${location}`, 400, reqRule);
          }

          // 6.2. Валидация (GET, Headers, Cookies) параметров
          if (reqRule.reqParams) {
            const paramsKeys = Object.keys(reqRule.reqParams);
            const params = {};

            paramsKeys.forEach((key) => {
              params[key] = ctx.request.query[key] || ctx.request.headers[key] || ctx.cookies.get(key) || undefined;
            });

            try {
              ctx.state.params = validateFields(Object.copy(params), reqRule.reqParams);
            } catch (err) {
              ErrorReq.throw(err, 400);
            }

            // reqlogger.debug("Params validated!", Object.keys( ctx.state.params || {} ) );
          } else {
            reqlogger.debug('Params didn\'t validating...');
            ctx.state.params = {};
          }

          // 6.3 Валидация BODY

          const reqBodyType = reqRule.reqType || 'json';
          let reqBodyStream;

          if ((ctx.method === 'POST' || ctx.method === 'PUT' || ctx.method === 'PATCH') && reqBodyType !== 'binary') {
            const opts = config.postBody || { encoding: 'utf8', limit: '250Mb' };

            if (reqBodyType !== 'stream') {
              // reqlogger.debug('bodyParse', opts );
              ctx.state.body = await bodyParse(ctx, opts);
            } else {
              const filePath = await bodyParse.text(ctx, opts);

              if (!filePath) {
                const err = new Error('MISS file path to stream');
                ErrorReq.throw(err, 400);
              }

              reqlogger.debug(`POST BODY stream from file "${filePath}"`);

              try {
                await fs.access(filePath, fs.constants.R_OK);
              } catch (err) {
                err.message = `${err.code === 'ENOENT' ? 'NOT FOUND file to stream' : 'ERROR read file system'}\n${err.message}`;
                ErrorReq.throw(err, 400);
              }

              reqBodyStream = fs.createReadStream(filePath);
              ctx.state.body = reqBodyStream;
            }
          }

          if (reqRule.reqBody) {
            reqlogger.debug('Validating body...');

            // reqlogger.debug("POST", ctx.state.body || {} )

            try {
              ctx.state.body = validateFields(ctx.state.body, reqRule.reqBody);
            } catch (err) {
              ErrorReq.throw(err, 400);
            }
          } else {
            reqlogger.debug('POST Body skip validation... reqRule.reqBody not found!', Object.keys(reqRule));
          }

          // -----------------------------------------------------------------------------------------------
          // 7. Формирование стейта

          ctx.state.sendRawResponseStatus = false;
          ctx.state.sendRawResponseData = false;

          ctx.state.mergeFrom(initSate);

          ctx.state.logger = reqlogger;
          ctx.state.config = config;

          ctx.state.reqError = ErrorReq;

          await authFieldsInit(ctx);

          // -----------------------------------------------------------------------------------------------
          // 8. Точка входа в кастомные мидллвары и команду

          await next();

          // -----------------------------------------------------------------------------------------------
          // 9. Запрос обработан

          // backProcesses
          if (reqBodyStream) {
            reqBodyStream.destroy();
            reqBodyStream = null;
          }

          ctx.status = ctx.state.sendRawResponseStatus || ctx.status || 200;

          ctx.set('rKey', reqlogger.rKey);

          if (config.response.json && !ctx.state.sendRawResponseData) {
            ctx.body = {
              code: ctx.status,
              data: ctx.body || {},
              srvTS: Date.now(),
              rKey: reqlogger.rKey
            };
          }

          if (ctx.state.sendRawResponseData) {
            ctx.body = ctx.body || `HTTP${ctx.status}`;
          }
        } catch (err) {
          // -----------------------------------------------------------------------------------------------
          // Отправляем  ошибку
          // -----------------------------------------------------------------------------------------------

          let status = err.status || err.statusCode || 500;

          if (typeof status !== 'number') {
            err.message = `${JSON.stringify(status)} ${err.message}`;
            status = 502;
          }

          ctx.status = status;

          ctx.set('err_code', status);
          ctx.set('rKey', reqlogger.rKey);

          ctx.body = {
            code: status,
            srvTS: Date.now(),
            rKey: reqlogger.rKey
          };

          if (config.response.errors) {
            ctx.res.setHeader('Content-Type', 'application/json');
            ctx.set('err_message', encodeURI(err.message));
            ctx.set('err_stack', encodeURI(err.stack));
            ctx.set('err_params', encodeURI(JSON.stringify(err.params)));

            ctx.body.message = err.message;
            ctx.body.stack = err.stack;
            ctx.body.params = err.params;
          }

          /**
           * Note: This is necessary for Nginx's auth_request
           */
          if (config.response.returnBodyInHeader) {
            ctx.set('err_body', JSON.stringify(ctx.body));
            ctx.body = {};
          }

          reqError = err;
        }

        // -----------------------------------------------------------------------------------------------
        // 10. Завершаем запрос

        // Заполняем "специальные" хедеры ответа из логгера
        helpers.updateResHeadersByLoggerEnv(ctx, reqlogger);

        const msg = `<- HTTP ${ctx.status} ${ctx.state.location} [ ${(ctx.response.length ? bytes(ctx.response.length, { decimalPlaces: 0 }) : '')} ]`;

        reqlogger.setEnv('status', ctx.status);

        reqlogger.info(msg);

        if (reqError && ctx.status !== 404) {
          const alertError = reqError.alert === undefined ? true : reqError.alert;
          reqlogger.error(reqError, ctx.state.params, alertError);

          if (!reqErrors[ctx.status]) {
            reqErrors[ctx.status] = 0;
          }

          reqErrors[ctx.status] += 1;
        }

        ctx.state.sendRawResponseStatus = false;
        ctx.state.sendRawResponseData = false;
      });

      // Загружаем кастомные мидлвары
      await loadMiddlewares(app);

      // Запускаем команду
      if (jsModules && Object.keys(jsModules).length > 0) {
        applogger.info('Add run command middleware...');

        app.use(async (ctx) => {
          const location = ctx.state.location;

          const cmd = jsModules[location];

          if (cmd) {
            ctx.state.logger.debug(`Run command ${location}`);
            await cmd(ctx);
          } else {
            SAError.throw('Command not loaded!', 500, { location });
          }
        });
      } else {
        applogger.info('No loaded commands...');
      }

      const server = app.listen(config.port);
      server.setTimeout(REQUEST_TIMEOUT);
      applogger.info(`⬆️ APP Start: ${config.port}`, null, true);

      server.on('error', (err) => {
        applogger.error(`Error on server starting: ${JSON.stringify(err)}`, null, true);
      });

      // GRACEFUL SHUTDOWN
      process.on('SIGINT', async () => {
        server.close(async () => {
          await end();
        });

        // Force close
        setTimeout(async () => {
          applogger.error('Forcing server close !!!', null, true);
          await end();
        }, 5000);
      });

      process.on('SIGTERM', async () => {
        server.close(async () => {
          await end();
        });

        // Force close
        setTimeout(async () => {
          applogger.error('Forcing server close !!!', null, true);
          await end();
        }, 5000);
      });

      return app;
    } catch (err) {
      // todo логгер может не создаться
      applogger.error(err, null, true);
      await end(err);
    }
  }

  //--------------------------------------------------------
  /**
   * Stop application
   *
   * @async
   * @param {*} err
   * @param {boolean} [killMePlease=true] Use process.kill
   */
  async function end(err, killMePlease = true) {
    try {
      await deps.end(initSate);
    } catch (e) {
      applogger.error(e);
    }

    applogger.info('⬇️App stopped!', null, true);

    if (killMePlease) {
      process.exit(err ? 1 : 0);
    }
  }

  //--------------------------------------------------------

  async function loadContract() {
    if (contract.reqParams) {
      contract.reqRules = contract.reqRules || {};
      contract.reqRules.reqParams = contract.reqRules.reqParams || {};
      contract.reqRules.reqParams = Object.merge(contract.reqRules.reqParams, contract.reqParams);

      delete contract.reqParams;
    }

    if (!contract.commands) {
      applogger.info('No Commands in contract...');

      return;
    }

    contract.reqRules = contract.reqRules || {};

    applogger.info(`LOAD commands from ${config.path.commands}`);

    const contractKeys = ['desc', 'reqMethod', 'reqType', 'reqParams', 'reqBody', 'default', 'reqAllows', 'reqNotAllows'];

    loadCommands(contract.commands, config.path.commands, '', contract.reqRules);

    const defKeys = Object.keys(defaultCmd);

    defKeys.forEach((defKey) => {
      defaultCmd[defKey] = flatDefaults(defKey);
    });

    function loadCommands(cmdObj, cmdPath, cmdKey = '', parentRules = {}, parentKey = '') {
      if (typeof cmdObj === 'string') {
        cmdObj = {
          desc: cmdObj
        };
      }

      const flatCmdKey = parentKey + ((cmdKey.length > 1) ? '/' : '') + cmdKey;
      let nextPath = cmdPath + ((cmdKey.length > 1) ? '/' : '') + cmdKey;

      nextPath = path.normalize(nextPath);

      const cmdProcessing = Object.copy(cmdObj);

      const rules = Object.copy(parentRules);

      contractKeys.forEach((contractKey) => {
        if (cmdObj.hasOwnProperty(contractKey)) {
          switch (contractKey) {
            case 'reqParams':
              rules[contractKey] = Object.merge(cmdObj[contractKey], parentRules[contractKey]);
              break;

            case 'reqAllows':
            case 'reqNotAllows':
              rules[contractKey] = parentRules[contractKey] ? [].concat(parentRules[contractKey]) : [];
              rules[contractKey].push(cmdObj[contractKey]);
              break;

            case 'default':
              defaultCmd[flatCmdKey.length > 0 ? flatCmdKey : '/'] = `${flatCmdKey}/${cmdObj[contractKey]}`;
              break;

            default:
              rules[contractKey] = cmdObj[contractKey];
          }

          delete cmdProcessing[contractKey];
        }
      });

      const comandsToProcess = Object.keys(cmdProcessing);

      if (comandsToProcess.length === 0) {
        // Это команда! нет "левых" ключей
        // delete rules[ "default" ];
        try {
          applogger.info(` --> ${flatCmdKey}`, { path: nextPath });
          jsModules[flatCmdKey] = require(nextPath);
          reqRules[flatCmdKey] = rules;
        } catch (e) {
          applogger.error(e, {
            cmdObj, cmdPath, cmdKey, parentRules, parentKey
          });
        }

        return true;
      }

      comandsToProcess.forEach((key) => {
        loadCommands(cmdObj[key], nextPath, key, rules, flatCmdKey);
      });

      return true;
    }

    function flatDefaults(defKey) {
      const defValue = defaultCmd[defKey];

      if (defaultCmd[defValue]) {
        return flatDefaults(defValue);
      }

      return defValue;
    }
  }

  //--------------------------------------------------------

  async function loadMiddlewares(currentApp) {
    if (!config.path.middlewares) {
      return;
    }

    let middlewares;

    try {
      middlewares = (await fs.readdir(config.path.middlewares)).sort();
    } catch (err) {
      applogger.debug(`No custom middlewares at: ${config.path.middlewares}`);

      return;
    }

    applogger.info(`LOAD middlewares at: ${config.path.middlewares}`);

    middlewares.forEach((file) => {
      try {
        currentApp.use(require(config.path.middlewares + file));
        applogger.info(`   + ${config.path.middlewares}${file}`);
      } catch (err) {
        SAError.throw(err, 500, { path: config.path.middlewares + file });
      }
    });
  }

  return { start, end };
};

//--------------------------------------------------------
/*
async function buildTest() {
  logger.info('TEST RUN FOR BUILD, CLOSE AFTER 10 SECONDS!');
  // TODO: надо тут че нить придумать - может по контракту обходить все и смотреть че как

  await asyncSleep(10);
  await end();
}
*/
//--------------------------------------------------------

module.exports = { getServer };
