/**
 * Created by Nikolay Ilin (na.iline@me.com) on 30.11.2018.
 * Copyrights by Soft-Artel LIMITED
 * www.soft-artel.com
 */

module.exports = { validateFields, validate, checkType };

// depricated: iso8601dt, date, float

// let rule = {
//
//     typeOf: [ 'guid',
//               'number',
//               'string',
//               'day',
//               'ts',
//               'dt',
//
//               'timeZone',
//               'phone',
//               'email',
//
//               'array',
//               'collection',
//               'object'
//
//             ],
//
//     dafault:    "value",
//     allowNull:  true,
//     between:    [ 0,1 ],
//     values:     [ "val1", "val2" ],
//     rules: {}
//
//      allow: 1,
//      shared: 1,
//      system: 1
//
// };

// Модели использования:
// 1) Проверяем DTO по набору правил
// 2) Проверяем отдельное поле по отдельному правилу

/** *
 * Validate fields by rules
 * @param fields
 * @param rules
 * @returns validFields || exception
 */

function validateFields(fields, rules, path = null, allErr = new Error('VALIDATE ERROR: ')) {
  try {
    allErr.code = 400;
    allErr.statusCode = allErr.code;
    allErr.status = allErr.code;

    // 0. Валидно ли правило
    if (!rules || typeof rules !== 'object') {
      // Сранная ситуация - правило не объект
      throwError(rules, 'Rules not valid!', path, fields);
    }

    // 1. Это правило для поля - надо проверять его
    if (rules.typeOf) {
      return validate(path, fields, rules);
    }
  } catch (error) {
    allErr.message += `\n ${error.message}`;
    throw (allErr);
  }

  // 2. Это вложенное правило - надо рекурсивно вызвать себя для каждого вложенного поля

  let isErrors = false;
  const validFields = {};

  fields = fields || {};

  for (const key in rules) {
    try {
      const subRule = rules[key];
      let subFields = fields[key];
      const subPath = path ? `${path}.${key}` : key;

      if (subRule.typeOf === undefined && subFields === undefined) {
        // Подправило тоже вложенное, а обьекта для него в полях нет
        subFields = {};
      }

      validFields[key] = validateFields(subFields, subRule, subPath, allErr);
    } catch (e) {
      isErrors = true;
      allErr.message = e.message;
    }
  }

  if (isErrors) {
    throw (allErr);
  }

  return validFields;
}

//-------------------------------------------------------------

/** *
 * Validate value by rule
 * @param key
 * @param value
 * @param rule
 * @param skipCollections - turn off check collections
 * @returns validValue || exception
 */

function validate(key, value, rule, skipCollections = false) {
  // 1. Предподготовка поля

  if ((rule.typeOf === 'number' || rule.typeOf === 'float' || rule.typeOf === 'ts') && typeof value === 'string') {
    // если требуется число, но пришла строка, привести к типу
    // eslint-disable-next-line no-restricted-globals
    value = isNaN(value * 1) ? value : value * 1;
  }

  if (value instanceof Date) {
    // если пришла Date, привести к типу
    switch (rule.typeOf) {
      case 'dt':
        value = value.toISOString();
        break;

      case 'ts':
        value *= 1;
        break;

      case 'day':
        value = value.toISOString().slice(0, 10).replace(/-/g, '') * 1;
        break;

      default:
        value = value.toString();
    }
  }

  // 1. Применяем дефолт есть он есть
  if (rule.default !== undefined && (value === rule.default || value === undefined || value === null)) {
    return rule.default;
  }

  // 2. Если дефолта нет - а валью не валидное - еррор
  if (value === undefined || value === null) {
    if (rule.allowNull) {
      return value;
    }

    throwError(rule, 'is undefined', key, value);
  }

  // 5. Проверяем тип поля

  if (!checkType(rule.typeOf, value)) {
    throwError(rule, 'invalid typeOf!', key, value);
  }

  // 6. Если есть values

  if (rule.values) {
    if (rule.values.indexOf(value) !== -1) {
      return value;
    }

    throwError(rule, 'value mismatch!', key, value);
  }

  // 7. between

  if (rule.between) {
    const min = rule.between[0] !== undefined ? rule.between[0] : 0;
    const max = rule.between[1] !== undefined ? rule.between[1] : min + 1;

    const val = (rule.typeOf === 'number') ? value * 1 : string(value).length;

    if (val >= min && val <= max) {
      return value;
    }

    throwError(rule, 'value mismatch!', key, value);
  }

  // 8. subrules
  if (rule.rules && skipCollections === false) {
    const allErr = new Error('');

    if (rule.typeOf === 'object') {
      return validateFields(value, rule.rules, key, allErr);
    }

    if (rule.typeOf === 'array') {
      let isErrors = false;
      const validArray = [];
      const subRule = rule.rules;

      for (let i = 0; i < value.length; i++) {
        try {
          let subFields = value[i];
          const subPath = `${key}[${i}]`;

          if (subRule.typeOf === undefined && subFields === undefined) {
            // Подправило тоже вложенное, а обьекта для него в полях нет
            subFields = {};
          }

          validArray.push(validateFields(subFields, subRule, subPath, allErr));
        } catch (error) {
          isErrors = true;
          allErr.message = error.message;
        }
      }

      if (isErrors) {
        throw (allErr);
      }

      return validArray;
    }

    if (rule.typeOf === 'collection') {
      let isErrors = false;
      const validCollection = {};
      const subRule = rule.rules;

      for (const subKey in value) {
        try {
          let subFields = value[subKey];
          const subPath = `${key}[${subKey}]`;

          if (subRule.typeOf === undefined && subFields === undefined) {
            // Подправило тоже вложенное, а обьекта для него в полях нет
            subFields = {};
          }

          validCollection[subKey] = validateFields(subFields, subRule, subPath, allErr);
        } catch (error) {
          isErrors = true;
          allErr.message = error.message;
        }
      }

      if (isErrors) {
        throw (allErr);
      }

      return validCollection;
    }

    throwError(rule, 'typeOf mismatch for subRules', key, value);
  }

  return value;
}

function checkType(type, value) {
  if (type === 'guid') {
    return typeof value === 'string' && match(value.toLowerCase(), /^[\da-f]{8}-[\da-f]{4}-[\da-f]{4}-[\da-f]{4}-[\da-f]{12}$/);
  }

  const jsTypes = ['number', 'string', 'boolean', 'object'];

  if (jsTypes.indexOf(type) !== -1) {
    // Тип дефалтовый
    // eslint-disable-next-line valid-typeof
    return typeof value === type;
  }

  if (type === 'array') {
    return Array.isArray(value);
  }

  if (type === 'collection') {
    return typeof value === 'object';
  }

  if (type === 'dt') {
    return match(value, /^\d{4}-\d{2}-\d{2}[ T]\d{2}:\d{2}:\d{2}/);
  }

  if (type === 'day') {
    return value > 19200101 && value < 30000101;
  }

  if (type === 'timeZone') {
    // eslint-disable-next-line no-useless-escape
    return match(value, /^(?:[a-z\d\-]+\/[a-z\d\-\+_]+[\/a-z\d\-\+_ ]*|GMT|UTC|)$/i);
  }

  if (type === 'phone') {
    return match(value, /^(\d{10,16})$/);
  }

  if (type === 'email') {
    // eslint-disable-next-line no-useless-escape
    return typeof value === 'string' && match(value.trim(), /^(([^<>()\[\]\\.,;:\s@"]+(\.[^<>()\[\]\\.,;:\s@"]+)*)|(".+"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$/);
  }

  if (type === 'ts') {
    return !Number.isNaN(Number(value));
  }

  return false;
}

function match(val, re) {
  return !!string(val).match(re);
}

function string(val) {
  return val === 0 ? '0' : `${val || ''}`;
}

function throwError(rule, msg, key = '', val = '') {
  // console.log( `${key} - ${msg} [${ JSON.stringify( val ) }] ${JSON.stringify(rule)}` );
  const err = new Error(`${key}:  ${msg} val:${JSON.stringify(val)} rule: ${JSON.stringify(rule)}`);
  throw err;
}
