/**
 * Created by Nikolay Ilin (na.iline@me.com) on 15.12.2017.
 * Copyrights by Soft-Artel LIMITED
 * www.soft-artel.com
 */

require('sa-prototype');

const fs = require('fs');
const Logger = require('./Logger');
const deps = require('./Dependencies');
const ErrorWrp = require('./Error');
const { asyncSleep } = require('./helpers');

const AppError = new ErrorWrp('Worker');

module.exports = (dirname) => {
  const pkg = require(`${dirname}/app.json`);

  let logger = null;

  const defaultProcessor = async (context) => {
    context.logger.info(`Stage : ${process.env.STAGE} not supported`);

    return true;
  };

  const config = {

    iterationSleep: 30, // seconds

    selfCheckInterval: 60 // seconds

  };

  let state;
  let workerLoop = true;
  const iterErrors = {};
  let iterationCount = 1;

  async function start(appConfig, processor, appState = {}) {
    try {
      config.mergeFrom(appConfig);
      logger = new Logger(
        {
          app: pkg.name,
          ver: pkg.version,
          stage: config.stage
        },
        Logger.TYPE.APPLICATION,
        config.loggerOptions || {}
      );
      state = appState;
      let usedProcessor = processor;

      state.config = config;

      logger.debug('starting...');

      const selfCheckResult = await deps.connect(state, config, logger);
      await selfCheck(selfCheckResult);
      let selfCheckDate = Date.now() + (config.selfCheckInterval * 1000);

      // GRACEFUL SHUTDOWN
      process.on('SIGINT', shutdownHandler);

      // GRACEFUL SHUTDOWN
      process.on('SIGTERM', shutdownHandler);

      logger.info('⬆️ APP Started', null, true);

      const runID = String.generateGUID();
      state.logger = logger.copy({}, Logger.TYPE.ITERATION);
      state.iterError = AppError;

      // use default processor if stage not supported
      if (config && config.stages) {
        if (!process.env.STAGE || !config.stages.includes(process.env.STAGE)) {
          config.iterationSleep = 300;
          usedProcessor = defaultProcessor;
        }
      }

      while (workerLoop) {
        state.logger.setEnv('rKey', `${runID}#${iterationCount}`);
        state.logger.debug(`Start #${iterationCount}`);
        state.logger.setEnv('rts', Date.now());

        try {
          await usedProcessor(state);
        } catch (error) {
          logger.error(error);
          error.status = error.status || 500;

          if (!iterErrors[error.status]) {
            iterErrors[error.status] = 0;
          }

          iterErrors[error.status] += 1;
        }

        if (selfCheckDate < Date.now()) {
          try {
            await selfCheck();
          } catch (err) {
            logger.error(err);
            await asyncSleep(15);
            await selfCheck();
          }

          selfCheckDate = Date.now() + (config.selfCheckInterval * 1000);
        }

        state.logger.debug(`End #${iterationCount}`, { sleep: config.iterationSleep, nextCheck: (new Date(selfCheckDate).toDtString()) });

        await asyncSleep(config.iterationSleep);

        iterationCount++;
      }

      await end();
    } catch (e) {
      await end(e);
    }
  }

  async function end(err) {
    workerLoop = false;

    await deps.end(state);
    // todo логгер может не создаться на старте
    logger.info('⬇️App stopped!', null, true);
    err && logger.error(err);

    process.exit(err ? 1 : 0);
  }

  async function selfCheck(selfCheckResult) {
    const checkState = selfCheckResult || await deps.check(state, iterErrors) || {};

    checkState.ts = Date.now();
    checkState.sleep = config.iterationSleep;
    checkState.iterCount = iterationCount;

    fs.writeFile('/tmp/_healthz.json', JSON.stringify(checkState),
      (error) => {
        if (error) { logger.error(`Health check file error - ${logger.getEnv('app')}`, error); }
      });
  }

  //--------------------------------------------------------

  function shutdownHandler() {
    workerLoop = false;

    // Force close
    setTimeout(async () => {
      logger.error('Forcing server close !!!');
      await end();
    }, 5000);
  }

  return { start, end, asyncSleep };
};
